@@ -1,8 +1,33 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #ifndef MUPDF_PDF_OBJECT_H
 #define MUPDF_PDF_OBJECT_H
 
+#include "mupdf/fitz/stream.h"
+
 typedef struct pdf_document pdf_document;
 typedef struct pdf_crypt pdf_crypt;
+typedef struct pdf_journal pdf_journal;
 
 /* Defined in PDF 1.7 according to Acrobat limit. */
 #define PDF_MAX_OBJECT_NUMBER 8388607
@@ -31,6 +56,7 @@ pdf_obj *pdf_new_array(fz_context *ctx, pdf_document *doc, int initialcap);
 pdf_obj *pdf_new_dict(fz_context *ctx, pdf_document *doc, int initialcap);
 pdf_obj *pdf_new_rect(fz_context *ctx, pdf_document *doc, fz_rect rect);
 pdf_obj *pdf_new_matrix(fz_context *ctx, pdf_document *doc, fz_matrix mtx);
+pdf_obj *pdf_new_date(fz_context *ctx, pdf_document *doc, int64_t time);
 pdf_obj *pdf_copy_array(fz_context *ctx, pdf_obj *array);
 pdf_obj *pdf_copy_dict(fz_context *ctx, pdf_obj *dict);
 pdf_obj *pdf_deep_copy_obj(fz_context *ctx, pdf_obj *obj);
@@ -54,15 +80,55 @@ int pdf_is_indirect(fz_context *ctx, pdf_obj *obj);
 */
 int pdf_obj_num_is_stream(fz_context *ctx, pdf_document *doc, int num);
 int pdf_is_stream(fz_context *ctx, pdf_obj *obj);
-pdf_obj *pdf_resolve_obj(fz_context *ctx, pdf_obj *a);
+
+/* Compare 2 objects. Returns 0 on match, non-zero on mismatch.
+ * Streams always mismatch.
+ */
 int pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b);
 int pdf_objcmp_resolve(fz_context *ctx, pdf_obj *a, pdf_obj *b);
+
+/* Compare 2 objects. Returns 0 on match, non-zero on mismatch.
+ * Stream contents are explicitly checked.
+ */
+int pdf_objcmp_deep(fz_context *ctx, pdf_obj *a, pdf_obj *b);
+
 int pdf_name_eq(fz_context *ctx, pdf_obj *a, pdf_obj *b);
 
 int pdf_obj_marked(fz_context *ctx, pdf_obj *obj);
 int pdf_mark_obj(fz_context *ctx, pdf_obj *obj);
 void pdf_unmark_obj(fz_context *ctx, pdf_obj *obj);
 
+typedef struct pdf_cycle_list pdf_cycle_list;
+struct pdf_cycle_list {
+	pdf_cycle_list *up;
+	int num;
+};
+int pdf_cycle(fz_context *ctx, pdf_cycle_list *here, pdf_cycle_list *prev, pdf_obj *obj);
+
+typedef struct
+{
+	int len;
+	unsigned char bits[1];
+} pdf_mark_bits;
+
+pdf_mark_bits *pdf_new_mark_bits(fz_context *ctx, pdf_document *doc);
+void pdf_drop_mark_bits(fz_context *ctx, pdf_mark_bits *marks);
+void pdf_mark_bits_reset(fz_context *ctx, pdf_mark_bits *marks);
+int pdf_mark_bits_set(fz_context *ctx, pdf_mark_bits *marks, pdf_obj *obj);
+
+typedef struct
+{
+	int len;
+	int max;
+	int *list;
+	int local_list[8];
+} pdf_mark_list;
+
+int pdf_mark_list_push(fz_context *ctx, pdf_mark_list *list, pdf_obj *obj);
+void pdf_mark_list_pop(fz_context *ctx, pdf_mark_list *list);
+void pdf_mark_list_init(fz_context *ctx, pdf_mark_list *list);
+void pdf_mark_list_free(fz_context *ctx, pdf_mark_list *list);
+
 void pdf_set_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int memo);
 int pdf_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int *memo);
 
@@ -105,6 +171,7 @@ pdf_obj *pdf_dict_geta(fz_context *ctx, pdf_obj *dict, pdf_obj *key, pdf_obj *ab
 pdf_obj *pdf_dict_gets(fz_context *ctx, pdf_obj *dict, const char *key);
 pdf_obj *pdf_dict_getsa(fz_context *ctx, pdf_obj *dict, const char *key, const char *abbrev);
 pdf_obj *pdf_dict_get_inheritable(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
+pdf_obj *pdf_dict_getp_inheritable(fz_context *ctx, pdf_obj *dict, const char *path);
 void pdf_dict_put(fz_context *ctx, pdf_obj *dict, pdf_obj *key, pdf_obj *val);
 void pdf_dict_put_drop(fz_context *ctx, pdf_obj *dict, pdf_obj *key, pdf_obj *val);
 void pdf_dict_get_put_drop(fz_context *ctx, pdf_obj *dict, pdf_obj *key, pdf_obj *val, pdf_obj **old_val);
@@ -126,18 +193,21 @@ void pdf_dict_put_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const cha
 void pdf_dict_put_text_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const char *x);
 void pdf_dict_put_rect(fz_context *ctx, pdf_obj *dict, pdf_obj *key, fz_rect x);
 void pdf_dict_put_matrix(fz_context *ctx, pdf_obj *dict, pdf_obj *key, fz_matrix x);
+void pdf_dict_put_date(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int64_t time);
 pdf_obj *pdf_dict_put_array(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int initial);
 pdf_obj *pdf_dict_put_dict(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int initial);
 pdf_obj *pdf_dict_puts_dict(fz_context *ctx, pdf_obj *dict, const char *key, int initial);
 
 int pdf_dict_get_bool(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 int pdf_dict_get_int(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
+int64_t pdf_dict_get_int64(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 float pdf_dict_get_real(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 const char *pdf_dict_get_name(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 const char *pdf_dict_get_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key, size_t *sizep);
 const char *pdf_dict_get_text_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 fz_rect pdf_dict_get_rect(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 fz_matrix pdf_dict_get_matrix(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
+int64_t pdf_dict_get_date(fz_context *ctx, pdf_obj *dict, pdf_obj *key);
 
 void pdf_array_push_bool(fz_context *ctx, pdf_obj *array, int x);
 void pdf_array_push_int(fz_context *ctx, pdf_obj *array, int64_t x);
@@ -201,9 +271,41 @@ char *pdf_load_stream_or_string_as_utf8(fz_context *ctx, pdf_obj *src);
 fz_quad pdf_to_quad(fz_context *ctx, pdf_obj *array, int offset);
 fz_rect pdf_to_rect(fz_context *ctx, pdf_obj *array);
 fz_matrix pdf_to_matrix(fz_context *ctx, pdf_obj *array);
+int64_t pdf_to_date(fz_context *ctx, pdf_obj *time);
 
+/*
+	pdf_get_indirect_document and pdf_get_bound_document are
+	now deprecated. Please do not use them in future. They will
+	be removed.
+
+	Please use pdf_pin_document instead.
+*/
 pdf_document *pdf_get_indirect_document(fz_context *ctx, pdf_obj *obj);
 pdf_document *pdf_get_bound_document(fz_context *ctx, pdf_obj *obj);
+
+/*
+	pdf_pin_document returns a new reference to the document
+	to which obj is bound. The caller is responsible for
+	dropping this reference once they have finished with it.
+
+	This is a replacement for pdf_get_indirect_document
+	and pdf_get_bound_document that are now deprecated. Those
+	returned a borrowed reference that did not need to be
+	dropped.
+
+	Note that this can validly return NULL in various cases:
+	1) When the object is of a simple type (such as a number
+	or a string), it contains no reference to the enclosing
+	document. 2) When the object has yet to be inserted into
+	a PDF document (such as during parsing). 3) And (in
+	future versions) when the document has been destroyed
+	but the object reference remains.
+
+	It is the caller's responsibility to deal with a NULL
+	return here.
+*/
+pdf_document *pdf_pin_document(fz_context *ctx, pdf_obj *obj);
+
 void pdf_set_int(fz_context *ctx, pdf_obj *obj, int64_t i);
 
 /* Voodoo to create PDF_NAME(Foo) macros from name-table.h */
@@ -225,6 +327,7 @@ enum {
 #define PDF_FALSE ((pdf_obj*)(intptr_t)PDF_ENUM_FALSE)
 #define PDF_LIMIT ((pdf_obj*)(intptr_t)PDF_ENUM_LIMIT)
 
+
 /* Implementation details: subject to change. */
 
 /*
@@ -232,4 +335,66 @@ enum {
 */
 void pdf_set_str_len(fz_context *ctx, pdf_obj *obj, size_t newlen);
 
+
+/* Journalling */
+
+/* Call this to enable journalling on a given document. */
+void pdf_enable_journal(fz_context *ctx, pdf_document *doc);
+
+/* Call this to start an operation. Undo/redo works at 'operation'
+ * granularity. Nested operations are all counted within the outermost
+ * operation. Any modification performed on a journalled PDF without an
+ * operation having been started will throw an error. */
+void pdf_begin_operation(fz_context *ctx, pdf_document *doc, const char *operation);
+
+/* Call this to start an implicit operation. Implicit operations are
+ * operations that happen as a consequence of things like updating
+ * an annotation. They get rolled into the previous operation, because
+ * they generally happen as a result of them. */
+void pdf_begin_implicit_operation(fz_context *ctx, pdf_document *doc);
+
+/* Call this to end an operation. */
+void pdf_end_operation(fz_context *ctx, pdf_document *doc);
+
+/* Call this to find out how many undo/redo steps there are, and the
+ * current position we are within those. 0 = original document,
+ * *steps = final edited version. */
+int pdf_undoredo_state(fz_context *ctx, pdf_document *doc, int *steps);
+
+/* Call this to find the title of the operation within the undo state. */
+const char *pdf_undoredo_step(fz_context *ctx, pdf_document *doc, int step);
+
+/* Helper functions to identify if we are in a state to be able to undo
+ * or redo. */
+int pdf_can_undo(fz_context *ctx, pdf_document *doc);
+int pdf_can_redo(fz_context *ctx, pdf_document *doc);
+
+/* Move backwards in the undo history. Throws an error if we are at the
+ * start. Any edits to the document at this point will discard all
+ * subsequent history. */
+void pdf_undo(fz_context *ctx, pdf_document *doc);
+
+/* Move forwards in the undo history. Throws an error if we are at the
+ * end. */
+void pdf_redo(fz_context *ctx, pdf_document *doc);
+
+/* Called to reset the entire history. This is called implicitly when
+ * a non-undoable change occurs (such as a pdf repair). */
+void pdf_discard_journal(fz_context *ctx, pdf_journal *journal);
+
+/* Internal destructor. */
+void pdf_drop_journal(fz_context *ctx, pdf_journal *journal);
+
+/* Internal call as part of saving a snapshot of a PDF document. */
+void pdf_serialise_journal(fz_context *ctx, pdf_document *doc, fz_output *out);
+
+/* Internal call as part of loading a snapshot of a PDF document. */
+void pdf_deserialise_journal(fz_context *ctx, pdf_document *doc, fz_stream *stm);
+
+/* Internal call as part of creating objects. */
+void pdf_add_journal_fragment(fz_context *ctx, pdf_document *doc, int parent, pdf_obj *copy, fz_buffer *copy_stream, int newobj);
+
+char *pdf_format_date(fz_context *ctx, int64_t time, char *s, size_t n);
+int64_t pdf_parse_date(fz_context *ctx, const char *s);
+
 #endif
